DEBRIS_POP_LIMIT = 16
MAX_COLLISION_GENS = 128
MAX_CYCLES = 4

from rlife import X, Pattern, GX2, GX
from rlife import EMPTY, GLIDER, BLOCK, BLINKER
from rlife import StampCollection

from inserters import AllConstructions, AllConstructions2

from collections import defaultdict, namedtuple

import time
import pprint
import os
import datetime
import string

def TimeStr():
    return datetime.datetime.now().time().isoformat()[:8]

class FileContext(object):
    contexts = []
    
    @classmethod
    def pollAll(cls):
        for c in cls.contexts:
            if c.poll(): return True
        return False
    
    @classmethod
    def pollFiles(cls):
        for c in cls.contexts:
            if c.pollFile(): return True
        return False
    
    def __init__(self, filename, seconds=None):
        self._filename = filename
        self._endTime = (None if seconds is None else
                         time.clock() + seconds)
    
    def __enter__(self):
        open(self._filename, "w").close()
        self.contexts.append(self)
        return self
    
    def __exit__(self, exceptionType, exceptionValue, traceback):
        if os.path.exists(self._filename):
            os.remove(self._filename)
        self.contexts.pop()
        return False
    
    def pollFile(self):
        return not os.path.exists(self._filename)

    def pollTime(self):
        return (self._endTime is not None and
                time.clock() >= self._endTime)

    def poll(self):
        return self.pollFile() or self.pollTime()
        
    def reason(self):
        return ("ran out of time"
                if os.path.exists(self._filename) else
                "cancelled by user")

def SafeFollowDelay(front, back):
    return 2 * (front.dbox[1] + front.dbox[3] - back.dbox[1]) + 8
    
def MinFollowDelays(sa, sb):
    def s0_then_s1(s0, s1):
        for t in range(SafeFollowDelay(s0, s1),
                       -SafeFollowDelay(s1, s0), -1):
            if (s0[t] | s1)[4] != s0[t + 4] | s1[4]:
                return t + 1
        return None
    ta = s0_then_s1(sa, sb)
    if ta is None:
        return 0, 1
    else:
        tb = s0_then_s1(sb, sa)
        return ta, tb

class _GGSeq(object):
    _delays = tuple([MinFollowDelays(GLIDER, GX2(i, 0))[0]
                     for i in range(8)])
    # http://conwaylife.com/forums/viewtopic.php?f=2&t=299
    assert _delays == (14, 14, 14, 13, 11, 9, 7, 0)
    
    def __getitem__(self, n):
        # The semiperimeter of the diagonal bounding box of the result,
        # measured in half-diagonals, is asymptotically equal to
        # (n/2)**0.5
        m = n // 2
        t = int(m ** 0.5)
        x = m - t * (t + 1)
        d = 2 * (t - abs(x)) + n % 2
        u = self._delays[min(abs(x), 7)] + (x >= 7)
        return GLIDER + GX2(x, u + d)

GGSeq = _GGSeq()
    
def TryHit(target, salvo):
    p = target + salvo
    t = 2 * p.dbox[3] + 4
    p = p[t]
    return None if p == target[t] | salvo[t] else (p, t)

def Collide(target, salvo, maxGens=128):
    skip = maxGens >> 3
    pt = TryHit(target, salvo)
    if pt is None: return None # Miss
    p, t = pt
    maxGens += t
    isFirsttime = True
    while True:
        if isFirsttime:
            isFirsttime = False
        else:
            if t + skip >= maxGens: return t, p # Timed out
            p = p[skip]
            t += skip
        p, n, rc = p.monitor(maxGens - t)
        t += n
        if rc <= 2: return t, p, None # Empty/p1/p2
        if rc == 5: return t, p # Timed out
        if rc == 3: continue # Possible p3
        # Possible p1/p2 with gliders
        center = EMPTY
        offspring = defaultdict(Pattern)
        try:
            for q in p.census(6):
                stride = (q[4] / q).xy # May raise ValueError
                if stride == (0, 0): # In center
                    center |= q
                elif stride[0] * stride[1] != 0 : # Glider/salvo
                    offspring[stride] |= q
                else: raise ValueError
        except ValueError: continue # Not center or glider
        if center[1] == center: center.setPeriod(1)
        elif center[2] == center: center.setPeriod(2)
        else: continue # Not p1/p2
        for s in offspring.itervalues(): s.setPeriod(4)
        return t, center, offspring

isAny = lambda x: True
isNone = lambda x: x is None
isNotNone = lambda x: x is not None
isEven = lambda x: x is not None and x & 1 == 0
isOdd = lambda x: x is not None and x & 1 != 0
isPositive = lambda x: x is not None and x > 0
isNegative = lambda x: x is not None and x < 0
isZero = lambda x: x is not None and x == 0

class ResultDict(defaultdict):
    def keysMatching(self, laneCond, dirCond, incCond):
        result = set()
        for k in self.iterkeys():
            if laneCond(k.lane) and dirCond(k.dir) and incCond(k.inc):
                result.add(k)
        return result

    def hasAny(self, laneCond, dirCond, incCond):
        for k in self.iterkeys():
            if laneCond(k.lane) and dirCond(k.dir) and incCond(k.inc):
                return True
        return False
    
    def isFeasible(self):
        return True
        
    def isComplete(self):
        return (self.hasAny(isNone, isNone, isPositive) and
                self.hasAny(isNone, isNone, isNegative) and
                (self.hasAny(isEven, isZero, isNotNone) and
                 self.hasAny(isOdd, isZero, isNotNone) or
                 self.hasAny(isEven, isPositive, isNotNone) and
                 self.hasAny(isOdd, isPositive, isNotNone)))

#    def essentialCleanups(self):
#        result = []
#        for laneCond, incCond in ((isNone, isPositive),
#                                  (isNone, isNegative),
#                                  (isEven, isAny),
#                                  (isOdd, isAny)):
#            if not self.keysMatching(laneCond, incCond, isZero):
#                result += [self.keysMatching(laneCond, incCond,
#                                                isPositive)]
#        return result
    
    def dirtyKeys(self):
        return self.keysMatching(isAny, isAny, isNone)

    def hasDirtyKeys(self):
        return self.hasAny(isAny, isAny, isNone)

class G(namedtuple("G", "x d")):
    def __getattr__(self, name):
        if name == "pattern":
            result = GLIDER[-self.d].translated(self.x, 0)
            self.pattern = result
        else: raise AttributeError, "Unknown attribute: " + name
        return result
    
    def __repr__(self):
        return "G" + tuple.__repr__(self)

class Salvo(object):
    def __init__(self, *args):
        self._contents = args
        
    def __getattr__(self, name):
        if name == "pattern":
            result = EMPTY
            for g in self._contents: result += g.pattern
            self.pattern = result
        else: raise AttributeError, "Unknown attribute: " + name
        return result
    
    def __repr__(self):
        return "Salvo" + str(self._contents).replace(",)", ")")
    
    def __eq__(self, other):
        return self._contents == other._contents
    
    def __hash__(self):
        return hash(self._contents)

class Cycle(object):
    
    @staticmethod
    def _f(salvo, delay):
        return EMPTY if delay is None else salvo.pattern[-delay]
    
    def __init__(self, *args):
        self._contents = args
        
    def __getattr__(self, name):
        if name == "pattern":
            result = sum(map(self._f, self.salvos, self._contents))
            self.pattern = result
        else: raise AttributeError, "Unknown attribute: " + name
        return result
    
    def __repr__(self):
        return "Cycle" + str(self._contents)

    def delayed(self, n):
        result = self.__class__(*((None if delay is None
                                   else delay + n)
                                  for delay in self._contents))
        result.setSalvos(self.salvos)
        return result
    
    def setSalvos(self, salvos):
        self.salvos = salvos
    
class Instruction(object):
    def __init__(self, *args):
        self._contents = args
        
    def __getattr__(self, name):
        if name == "pattern":
            result = self.pattern = sum(c.pattern
                                        for c in self._contents)
        elif name == "elbowPattern":
            d = SafeFollowDelay(self.elbow, self.pattern)
            result = self.safePattern = self.elbow + self.pattern[-d]
        else: raise AttributeError, "Unknown attribute: " + name
        return result

    def __repr__(self):
        return "Instruction" + str(self._contents).replace(",)", ")")
    
    def __len__(self):
        return len(self._contents)
    
    def __getitem__(self, i):
        return self._contents[i]
    
    def delayed(self, d):
        return self.__class__(*(cycle.delayed(d)
                              for cycle in self._contents))
        
    def quickstr(self):
        return string.ascii_uppercase[len(self._contents) - 1]

class Experiment(object):
    def __init__(self, elbow, salvo, target):
        self.elbow, self.salvo = elbow, salvo
        #self.inDebris = inDebris
        self.target = target
        target.period
        
    def run(self, maxGens=MAX_COLLISION_GENS):
        d = SafeFollowDelay(self.target, self.salvo)
        tco = Collide(self.target, self.salvo[-d], maxGens)
        self.timedOut = len(tco) == 2
        if self.timedOut:
            self.time, self.center = tco
            return False # Timed out
        self.time, self.center, self.offspring = tco
        if self.center.pop == 0:
            return False # No center
        if self.center.pop > DEBRIS_POP_LIMIT + self.elbow.pop:
            return False # Population too great
        self.inc = None
        if self.center.pop == self.elbow.pop:
            try:
                xform = self.center / self.elbow # May raise ValueError
                if not xform.isReoriented: # Is translated elbow
                    x, y = xform.xy
                    if x == y: # Is translated along diagonal
                        self.inc = -x
            except ValueError: pass
        # Find the output lane, if any
        if self.offspring is not None:
            if len(self.offspring) != 1:
                return False # Too many offspring
            if ((-1, 1) not in self.offspring and
                (1, -1) not in self.offspring):
                return False # Offspring not SW or NE
            g = self.offspring.values()[0]
            if g.pop != 5:
                return False # Output not a glider
            self.lane = g.ld[0]
            self.dir = 1 if (1, -1) in self.offspring else 0
        else:
            self.lane = self.dir = None
        # Success
        return True

ALL_CONVOY_PERMS = ((0,), (1,), (2,),
                    (0, 1), (0, 2), (1, 0), (1, 2), (2, 0), (2, 1),
                    (0, 1, 2), (0, 2, 1), (1, 0, 2),
                    (1, 2, 0), (2, 0, 1), (2, 1, 0))

class Operation(namedtuple("Operation", "lane dir inc debris")):
    def __new__(cls, lane=None, dir=None, inc=None, debris=EMPTY):
        return tuple.__new__(cls, (lane, dir, inc, debris))
        
    def __getattr__(self, name):
        if name == "pop":
            result = self.pop = self.debris.pop
        elif name == "fileStr":
            result = self.fileStr = ("(lane=" + str(self.lane) +
                                     ", dir=" + str(self.dir) +
                                     ", inc=" + str(self.inc) +
                                     ", pop=" + str(self.pop) + ")")
        else: raise AttributeError, "Unknown attribute: " + name
        return result
    
    def repr(self):
        s = self.__repr__()
        return (s
                .replace("lane=None, dir=None, ", "")
                .replace("inc=None, ", "")
                .replace(", debris=P.EMPTY", ""))

    def quickStr(self, ins):
        insLetter = string.ascii_uppercase[len(ins._contents) - 1]
        result = str(self.inc) + insLetter
        if self.lane is not None:
            result = str(self.lane) + "@" + result
        return result
    
class Assessor(object):
    def __init__(self, elbow, salvos, searches,
                 essentialMinutes, bonusMinutes,
                 forceCompleteASet=False):
        Instruction.elbow = elbow
        self.elbow, self.salvos = elbow, salvos
        self.searches = searches
        self.essentialSeconds = essentialMinutes * 60
        self.bonusSeconds = bonusMinutes * 60
        self.forceCompleteASet = forceCompleteASet
        self.isUpdated = False
        self.startTime = time.clock()
        
    def dump(self):
        return
        t = time.clock() - self.startTime
        for isClean in [False, True]:
            print "Clean results" if isClean else "Dirty results"
            for isMove in [True, False]:
                if isMove:
                    print "  Movement"
                    if isClean: print "    Inc  Count"
                    else: print "    Inc  Count  Debris"
                else:
                    print "  Output"
                    if isClean: print "    Lane  Inc  Count"
                    else: print "    Lane  Inc  Count  Debris"
                lastKey = None
                for k in sorted(self.results.iterkeys()):
                    if ((k.pop == 0) == isClean and
                            (k.lane is None) == isMove):
                        debstr = ("" if isClean else
                               "%2d p%1d %s" %
                               (k.pop, k.debris.period,
                                (self.elbow + k.debris).homed.rle))
                        data = (len(self.results[k]), debstr)
                        if isMove:
                            if k[:2] == lastKey:
                                print "       %7d  %s" % data
                            else:
                                print ("%7d%7d  %s" %
                                       ((k.inc,) + data))
                        else:
                            if k[:2] == lastKey:
                                print "             %7d  %s" % data
                            else:
                                print "%8d%5d%7d  %s" % (k[:2] + data)
                        lastKey = k[:2]
                print
        print "%7d experiments performed" % self.experimentCount
        print "%7d results assessed" % self.assessCount
        print "%7d results retained" % self.retainCount
        print "%6.0fs execution time" % t
        
    def search(self, op):
        
        def findMaxDelay(target, salvoA, salvoB, cachekey):
            try:
                maxDelay = maxDelays[cachekey]
            except KeyError:
                exp = Experiment(self.elbow, salvoA.pattern, target)
                ok = exp.run()
                t = exp.time
                if op is not None and exp.timedOut:
                    p = target | salvoA.pattern
                    n = p.pop + DEBRIS_POP_LIMIT
                    for i in xrange(0, t, 8):
                        if p[i].pop > n:
                            #print "reduced", t, i
                            t = i
                            break
                maxDelay = t + SafeFollowDelay(exp.center,
                                               salvoB.pattern)
                maxDelays[cachekey] = maxDelay
            return maxDelay
        
        def makeCycle(perm, delays):
            cycle = [None, None, None]
            for i, d in zip(perm, (0,) + delays): cycle[i] = d
            result = Cycle(*cycle)
            result.setSalvos(self.salvos)
            return result
        
        self.currentResults = ResultDict(list)
        target = self.elbow if op is None else op.debris
        safeDelay = max(*((SafeFollowDelay(target, s.pattern)
                          for s in self.salvos)))
        perms = ALL_CONVOY_PERMS
        perms = ((0,), (1,), (0, 1), (1, 0))
        for salvo in self.salvos:
            if TryHit(target, salvo.pattern[-safeDelay]) is None:
                perms = filter(lambda p:
                               self.salvos[p[0]] is not salvo, perms)
        minDelays = dict()
        #for pair in ((0, 1), (0, 2), (1, 2)):
        for pair in ((0, 1),):
            salvos = [self.salvos[i] for i in pair]
            m, n = MinFollowDelays(salvos[0].pattern,
                                   salvos[1].pattern)
            minDelays[pair] = m
            minDelays[tuple(reversed(pair))] = n
        maxDelays = dict()
        for perm in perms:
            permSalvos = [self.salvos[i] for i in perm]
            if len(perm) == 1:
                ins = Instruction(makeCycle(perm, ()))
                self.add(ins, op)
                if FileContext.pollFiles(): return False
            elif len(perm) == 2:
                salvoA, salvoB = permSalvos
                minDelay = max(minDelays[perm], 0)
                maxDelay = findMaxDelay(target, salvoA, salvoB, perm)
                #for delay in range(minDelay, maxDelay):
                for delay in range(maxDelay - 1, minDelay - 1, -1):
                    ins = Instruction(makeCycle(perm, (delay,)))
                    self.add(ins, op)
                    if FileContext.pollFiles(): return False
            elif len(perm) == 3:
                salvoA, salvoB, salvoC = permSalvos
                minDelayAB = max(minDelays[perm[:2]], 0)
                minDelayBC = minDelays[perm[1:]]
                minDelayAC = max(minDelays[perm[0], perm[2]], 0)
                maxDelayAB = findMaxDelay(target, salvoA, salvoB,
                                            perm[:2])
                #for delayAB in range(minDelayAB, maxDelayAB):
                for delayAB in range(maxDelayAB - 1,
                                     minDelayAB - 1, -1):
                    salvoAB = makeCycle(perm[:2], (delayAB,))
                    minDelayBCmax = max(delayAB + minDelayBC,
                                        minDelayAC)
                    maxDelayBC = findMaxDelay(target, salvoAB, salvoC,
                                              (perm, delayAB))
                    #for delayBC in range(minDelayBCmax, maxDelayBC):
                    for delayBC in range(maxDelayBC - 1,
                                         minDelayBCmax - 1, -1):
                        ins = Instruction(makeCycle(perm,
                                                 (delayAB, delayBC)))
                        self.add(ins, op)
                        if FileContext.pollFiles(): return False
        if self.currentResults:
            for k, v in self.currentResults.iteritems():
                self.results[k] += v
                self.retainCount += len(v)
            self.isUpdated = True
        else:
            if op:
                print TimeStr(), "No cleanups found"
        self.currentResults = None
        return True
    
    def add(self, ins, op):
        self.add2(ins, op)
        if op is not None and op.debris.period == 2:
            self.add2(ins.delayed(1), op)
    
    def add2(self, ins, op):
        target = op.debris if op is not None else self.elbow
        exp = Experiment(self.elbow, ins.pattern, target)
        self.experimentCount += 1
        if not exp.run(): return
        self.assessCount += 1
        lane, dir = ((op.lane, op.dir)
                     if op is not None and exp.lane is None
                     else (exp.lane, exp.dir))
        if lane is None and exp.inc is 0: return # Nop
        debris = (exp.center.canonicalPhase if exp.inc is None
                  else EMPTY)
        if op is None:
            newop = Operation(exp.lane, exp.dir, exp.inc, debris)
            self.currentResults[newop] += [ins]
            self.retainCount += 1
            self.isUpdated = True
            return
#        if exp.inc is None: # Left debris
#            newop = Operation(exp.lane, exp.dir, None,
#                            exp.center.canonicalPhase)
#            self.currentResults[newop] += [ins]
#            self.retainCount += 1
#            self.isUpdated = True
#            return
        if op.lane is not None and exp.lane is not None:
            return # Two outputs
        lane, dir = ((op.lane, op.dir) if exp.lane is None
                     else (exp.lane, exp.dir))
        newop = Operation(lane, dir, exp.inc, debris)
        isNew = newop not in self.currentResults
        isFound = False
        for ains in self.results[op]:
            if (exp.inc is None and
                len(ains._contents) == MAX_CYCLES - 1):
                continue
            isFound = isFound or exp.inc is not None
            #d = SafeFollowDelay(ains.pattern, ins.pattern)
            exp = Experiment(self.elbow, ains.pattern, self.elbow)
            exp.run(10000)
            d = 512
            if exp.center != op.debris:
                if exp.center != op.debris[1]:
                    print ains.elbowPattern.homed.rle
                    print ins.elbowPattern.homed.rle
                assert exp.center == op.debris[1]
                d += 1
            cycles = ains._contents + (ins[0].delayed(exp.time + d),)
            newins = Instruction(*cycles)
            self.currentResults[newop] += [newins]
#            exp1 = Experiment(self.elbow, ins.pattern, op.debris)
#            exp2 = Experiment(self.elbow, newins.pattern, self.elbow)
#            exp1.run(10000)
#            exp2.run(10000)
#            if exp1.center.canonicalPhase != exp2.center.canonicalPhase:
#                print (ins.pattern[-SafeFollowDelay(op.debris, ins.pattern)] + op.debris).homed.rle
#                print newins.elbowPattern.homed.rle
#                print exp.time, d
#                assert False
        if isNew and isFound:
            print TimeStr(), ("Cleanup(s) found: " + newop.repr() +
                              " - for example:")
            print TimeStr(), "  " + newins.elbowPattern.homed.rle

    def clean(self, keys, subset, isEssential):
        op = sorted(subset, key=lambda k: k.pop)[0]
        print
        print TimeStr(), "Looking for cleanups of " + op.repr()
        print TimeStr(), ("  " +
                          self.results[op][0].elbowPattern.homed.rle)
        ext = ".essential" if isEssential else ".bonus"
        with FileContext(op.fileStr + ext):
            if self.search(op):
                self.retainCount -= len(self.results[op])
                del self.results[op]
                self.isUpdated = True
            else:
                print TimeStr(), "Cleanup search aborted but retained"
        keys.remove(op)

    def run(self):
        self.results = ResultDict(list)
        s = self.searches.get((self.elbow, self.salvos))
        if s is not None:
            if len(s) == 0:
                return
            instructions = s["instructions"]
            for op in instructions:
                op.debris.divine(2)
            self.results.update(instructions)
            if len(self.results) == 0:
                return
            for instructions in self.results.itervalues():
                for instruction in instructions:
                    for cycle in instruction._contents:
                        cycle.setSalvos(self.salvos)
            self.experimentCount = s["experimentCount"]
            self.assessCount = s["assessCount"]
            self.retainCount = s["retainCount"]
            print TimeStr(), "Resuming search:", self.salvos
            print
        else:
            self.experimentCount = 0
            self.assessCount = 0
            self.retainCount = 0
            self.isUpdated = True
            print TimeStr(), "Starting search:", self.salvos
            print
            self.search(None)
        self.dump()
#        if (not self.results.isComplete() and
#                self.results.isFeasible()):
#            print
#            print TimeStr(), "No complete operation set"
#            print TimeStr(), "Searching for missing operations"
#            print
#            with FileContext("ALL.essential",
#                             self.essentialSeconds) as fc:
#                keys = set().union(*self.results.essentialCleanups())
#                while (keys and not self.results.isComplete() and
#                       self.results.isFeasible()):
#                    if FileContext.pollAll():
#                        print
#                        print TimeStr(), ("Search for missing " +
#                                          "operations " + fc.reason())
#                        break
#                    subset = sorted(self.results.essentialCleanups(),
#                                    key=lambda x: len(x))[0] & keys
#                    self.clean(keys, subset, True)
#        print
#        if self.forceCompleteASet and not self.results.isComplete():
#            if not self.results.isFeasible():
#                print TimeStr(), ("No complete operation set is "
#                                  "possible - abandoning search")
#                self.results = ResultDict(list)
        if False:
            pass
        else:
            if self.results.isComplete():
                print TimeStr(), "Complete operation set found"
                print TimeStr(), "Searching for bonus operations"
            else:
                print TimeStr(), "No complete operation set found"
                print TimeStr(), "Forced search for bonus operations"
            with FileContext("ALL.bonus", self.bonusSeconds) as fc:
                while True:
                    keys = self.results.dirtyKeys()
                    if not keys: break
                    while keys:
                        if FileContext.pollAll():
                            print
                            print TimeStr(),
                            print ("Search for bonus operations " +
                                   fc.reason())
                            break
                        self.clean(keys, keys, False)
        #self.store()
        if self.isUpdated:
            if not self.results.hasDirtyKeys():
                if not self.results.isComplete():
                    self.results = ResultDict(list)
            print
            print TimeStr(), "Updating results.py ...",
            with open("results.py", "a") as file:
                self.store(file)
            print "done"
        print
        print TimeStr(), "Search run completed"
        return self.results
        
    def store(self, file=None):
        if file is None:
            print>>file, ("####################################"
                          "####################################")
        lanes = tuple(s._contents[0].x for s in self.salvos)
        print>>file, ("Searches[" + str(self.elbow) + ", " +
                      "salvos" + str(lanes) + "] ="),
        if len(self.results) == 0:
            print>>file, "{}"
        else:
            print>>file, "dict("
            print>>file
            print>>file, "  instructions = {"
            firstTime = True
            for isClean in (True, False):
                commentPrinted = False
                for k, v in sorted(self.results.iteritems()):
                    if (k.pop == 0) == isClean:
                        if not firstTime: print>>file, ","
                        if not commentPrinted:
                            print>>file
                            s = ("Complete operations" if isClean else
                                 "Possible operations not yet "
                                 "cleared of debris")
                            print>>file, "      # " + s
                            print>>file
                            commentPrinted = True
                        print>>file, "    " + k.repr() + ": ["
                        if isClean:
                            p = "s" if len(v) > 1 else ""
                            s = str(len(v[0])) + "-cycle "
                            if k.lane is None:
                                s += "INC" if k.inc > 0 else "DEC"
                                s += str(abs(k.inc)) + p
                            else:
                                s += "ODD" if k.lane % 2 else "EVEN"
                                s += p + " on " + str(k.lane)
                            s += ":"
                            print>>file, "        # " + s
                        firstTime2 = True
                        for r in v:
                            if not firstTime2: print>>file, ","
                            print>>file, ("        # " +
                                          r.elbowPattern.homed.rle)
                            s = str(r).replace("None", "na")
                            print>>file, "      " + s,
                            firstTime2 = False
                        print>>file, "]",
                        firstTime = False
            print>>file, "} ,"
            print>>file
            print>>file, ("  experimentCount = " +
                          str(self.experimentCount) + ",")
            print>>file, ("  assessCount = " +
                          str(self.assessCount) + ",")
            print>>file, ("  retainCount = " +
                          str(self.retainCount) + " )")
        print>>file, ("####################################"
                      "####################################")
        
def Foo(lanes, searches):
    salvos = tuple(map(lambda x: Salvo(G(x, 0)), lanes))
    s = searches[BLOCK, salvos]
    if not s: return None
    results = s["instructions"]
    return AllConstructions2(salvos, results, ResultDict)
