'''
Created on 3 Sep 2009

@author: Paul Chapman
'''

import collections
import rle
import _ctypes

from ctypes import *
from fractions import gcd

class Xform(tuple):
    def __new__(cls, *args):
        if len(args) == 1:
            r = (0, 0, 1, 0, 0, 1) + args
        elif len(args) == 2:
            r = args + (1, 0, 0, 1, 0)
        elif len(args) == 4:
            r = (0, 0) + args + (0,)
        elif len(args) == 7:
            r = args
        else:
            raise TypeError
        return tuple.__new__(cls, r)

    def __getattr__(self, name):
        if name == "x": result = self.x = self[0]
        elif name == "y": result = self.y = self[1]
        elif name == "xy": result = self.xy = self[:2]
        elif name == "t": result = self.t = self[6]
        elif name == "isReoriented":
            result = self.isReoriented = self[2:6] != (1, 0, 0, 1)
        else: raise AttributeError, "Unknown attribute: " + name
        return result

    def __repr__(self):
        m = self.__class__(*self[2:6])
        for name in ["I", "R", "C", "A", "V", "H", "D", "S"]:
            if m == self.__class__.__dict__[name]:
                sm = 'X.' + name
                break
        sxy = 'X' + str(self.xy)
        st = 'X(' + str(self.t) + ')'
        if self.x == self.y == self.t == 0: return sm
        if m == self.I:
            if self.t == 0: return sxy
            if self.x == self.y == 0: return st
            return sxy + '*' + st
        else:
            if self.t == 0: return sxy + '*' + sm
            if self.x == self.y == 0: return sm + '*' + st
            return sxy + '*' + sm + '*' + st
    
    def __invert__(self):
        x, y, a, b, c, d, t = self
        p, q, r, s = ((d, -b, -c, a) if a * d > b * c
                      else (-d, b, c, -a))
        z, w = -(p * x + q * y), -(r * x + s * y)
        return self.__class__(z, w, p, q, r, s, -t)
    
    def __call__(self, other):
        if isinstance(other, collections.Sequence) and len(other) == 2:
            x, y, a, b, c, d, t = self
            z, w = other
            if t != 0:
                raise TypeError, "Cannot temporally transform a point"
            return (x + a * z + b * w,
                    y + c * z + d * w)
        raise ValueError, "Can only apply an Xform to (x, y)"
    
    def __mul__(self, other):
        if isinstance(other, Xform):
            x, y, a, b, c, d, t = self
            z, w, p, q, r, s, u = other
            return self.__class__(x + a * z + b * w,
                                  y + c * z + d * w,
                                  a * p + b * r,
                                  a * q + b * s,
                                  c * p + d * r,
                                  c * q + d * s,
                                  t + u)
        return NotImplemented
    
    def __div__(self, other):
        if isinstance(other, Xform):
            return self.__mul__(other.__invert__())
        return NotImplemented
    
    def split(self):
        return (self.__class__(*self[:2]),
                self.__class__(*self[2:6]),
                self.__class__(self[6]))

X = Xform

XI = X.I = X( 1,  0,  0,  1)
XR = X.R = X(-1,  0,  0, -1)
XC = X.C = X( 0, -1,  1,  0)
XA = X.A = X( 0,  1, -1,  0)
XV = X.V = X(-1,  0,  0,  1)
XH = X.H = X( 1,  0,  0, -1)
XD = X.D = X( 0,  1,  1,  0)
XS = X.S = X( 0, -1, -1,  0)

X.ORIENTATIONS = (XI, XR, XC, XA, XV, XH, XD, XS)

_lib = windll.RLife8
    
class _BoundingBoxes(Structure):
    _fields_ = [("w", c_int),
                ("n", c_int),
                ("e", c_int),
                ("s", c_int),
                ("sw", c_int),
                ("nw", c_int),
                ("ne", c_int),
                ("se", c_int)]

class _Header(Structure):
    _fields_ = [("size", c_int),
                ("hash", c_int),
                ("pop", c_int),
                ("period", c_int),
                ("boxes", _BoundingBoxes),
                ("xstride", c_int),
                ("ystride", c_int),
                ("jogged", c_ubyte),
                ("normalizer", c_ubyte),
                ("symmetryGroup", c_ubyte),
                ("flags", c_ubyte)]

_NORMALIZER_MASK = 0x01
_SYMMETRYGROUP_MASK = 0x02
_POPULATION_MASK = 0x04
_BOUNDINGBOX_MASK = 0x08
_PERIODIC_MASK = 0x10
_HASH_MASK = 0x20

class Pattern(object):

    class Type(object):
        EMPTY = 0
        P1 = 1
        P2 = 2
        MAYBE_P3_WITH_SHIPS = 3
        MAYBE_SHIPS = 4
        TIME_OUT = 5

    _buffer = (c_int * 16000)()
    _smearBuffer = (c_int * 16000)()
    
    @staticmethod
    def _trim(buffer):
        n = buffer[0] / 4
        return (c_int * n)(*buffer[:n])

    @staticmethod
    def _factors(n):
        l = reduce(list.__add__, ([i, n / i] for i in
                                  range(1, int(n ** 0.5) + 1)
                                  if n % i == 0))
        return sorted(list(set(l)))
    
    @classmethod
    def _wrapbuffer(cls):
        return cls(cls._trim(cls._buffer))
    
    @classmethod
    def _wrapsmear(cls):
        return cls(cls._trim(cls._smearBuffer))
    
    @classmethod
    def _build(cls, seq):
        if len(seq) == 0:
            return cls()
        elif len(seq) == 2:
            return cls(*seq)
        elif (len(seq) & 1) == 0:
            n = (len(seq) >> 1) & -2
            return cls(seq[:n]) | cls(seq[n:])
        else:
            raise TypeError, "Cell-list length must be even"

    def __init__(self, *args):
        if len(args) == 0:
            _lib.Empty(self._buffer)
            self._contents = self._trim(self._buffer)
            self.setPeriod(1, 0, 0, verify=False)
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, str):
                self._contents = self._build(rle.cells(arg))._contents
            elif isinstance(arg, _ctypes.Array):
                self._contents = arg
            elif isinstance(arg, collections.Sequence):
                self._contents = self._build(arg)._contents
            else:
                raise TypeError, ("Pattern(x): x must be a cell list, "
                                  "string or ctypes Array")
        elif len(args) == 2:
            _lib.SingleCell(self._buffer, c_int(args[0]),
                            c_int(args[1]))
            self._contents = self._trim(self._buffer)
        else:
            self._contents = self._build(args)._contents
        self._phases = dict()
        self._phases[0] = self

    def __getattr__(self, name):
        if name == "_header":
            result = self._header = cast(self._contents,
                                         POINTER(_Header))[0]
        elif name == "_unjogged":
            result = self._unjogged = (self._jogged()
                                       if self._header.jogged
                                       else self)
        elif name == "_rawData":
            result = self._rawData = self._unjogged._contents[15:]
        elif name == "pop":
            result = self.pop =_lib.Population(self._contents)
        elif name == "hash":
            result = self.hash = _lib.Hash(self._contents)
        elif name == "cells":
            _lib.Points(self._buffer, self._contents)
            result = self.cells = list(self._buffer[:self.pop * 2])
        elif name == "firstcell":
            if _lib.FirstPoint(self._buffer, self._contents):
                raise IndexError, "Pattern is empty"
            result = self.firstcell = list(self._buffer[:2])
        elif name == "box":
            boxes = self._boxes()
            result = self.box = (boxes.w,
                                 boxes.n,
                                 boxes.e - boxes.w,
                                 boxes.s - boxes.n)
        elif name == "dbox":
            boxes = self._boxes()
            result = self.dbox = (boxes.sw,
                                  boxes.nw,
                                  boxes.ne - boxes.sw,
                                  boxes.se - boxes.nw)
        elif name == "rle": 
            x, y = self.box[:2]
            if x < 0 or y < 0:
                raise ValueError, "Pattern is not in first quadrant"
            result = self.rle = rle.rle(self.cells)
        elif name == "homer":
            x, y = self.box[:2]
            result = self.homer = X(-x, -y)
        elif name == "isPeriodic":
            return self._header.flags & _PERIODIC_MASK != 0
        elif name == "_pst":
            if self._header.flags & _PERIODIC_MASK == 0:
                raise ValueError, "Pattern is not known to be periodic"
            result = self._pst = (self._header.period,
                                  self._header.xstride,
                                  self._header.ystride)
        elif name == "period":
            if self._header.flags & _PERIODIC_MASK == 0:
                raise ValueError, "Pattern is not known to be periodic"
            result = self.period = self._header.period
        elif name == "stride":
            if self._header.flags & _PERIODIC_MASK == 0:
                raise ValueError, "Pattern is not known to be periodic"
            result = self.stride = (self._header.xstride,
                                    self._header.ystride)
        elif name == "homed":
            result = self.homed = self.homer * self
#            if self._header.flags & _PERIODIC_MASK != 0:
#                self.homed = result
        elif name == "ld":
            z = (self / G).split()
            x, y = (~z[1] * z[0] * z[1]).xy
            result = self.ld = x - y, 4 * y - z[2].t
            #assert self == z[1] * X(result[0], 0) * G[-result[1]]
        elif name == "canonicalPhase":
            if self.stride != (0, 0):
                raise ValueError, "Pattern is not an oscillator"
            result = min((self[i] for i in range(self.period)))
            self.canonicalPhase = result
        else:
            raise AttributeError, "Unknown attribute: " + name
        return result
    
    def __getitem__(self, i):
        if i == 0:
            return self
        xlate = (0, 0)
        if self.isPeriodic:
            n = self.period
            q, i = divmod(i, n)
            x, y = self.stride
            xlate = (q * x, q * y)
        elif i < 0:
            raise IndexError, ("Cannot devolve pattern "
                               "not known to be periodic")
        if self._phases.has_key(i):
            result = self._phases[i]
        else:
            m = 0
            for k in self._phases.keys():
               if m < k < i: m = k
            p = self[m]
            _lib.Gen(self._buffer, p._contents, None, c_int(i - m),
                     None, None)
            result = self._phases[i] = self._wrapbuffer()
        if not result.isPeriodic and self.isPeriodic:
            result.setPeriod(*self._pst, verify=False)
        return result.translated(*xlate)
        
    def __hash__(self):
        return self.hash
    
    def __nonzero__(self):
        return self.pop != 0

    def __repr__(self):
        for name in self.PATTERNS:
            p = self.__class__.__dict__[name]
            try:
                xform = self / p
                t = xform.t
                xform /= X(t)
                st = '' if t == 0 else str([t])
                sx = '' if xform == XI else str(xform) + '*'
                return sx + 'P.' + name + st
            except ValueError: pass
        try:
            s = 'P("' + self.rle + '")'
        except ValueError:
            s = str(~self.homer) + '*P("' + self.homed.rle + '")'
        return s
    
    def __and__(self, other):
        if isinstance(other, self.__class__):
            _lib.And(self._buffer, self._contents, other._contents)
            return self._wrapbuffer()
        return NotImplemented

    def __or__(self, other):
        if not other:
            return self
        if isinstance(other, self.__class__):
            _lib.Or(self._buffer, self._contents, other._contents)
            return self._wrapbuffer()
        return NotImplemented

    def __xor__(self, other):
        if isinstance(other, self.__class__):
            _lib.Xor(self._buffer, self._contents, other._contents)
            return self._wrapbuffer()
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, self.__class__):
            _lib.Andnot(self._buffer, self._contents, other._contents)
            return self._wrapbuffer()
        return NotImplemented

    def __add__(self, other):
        if not other:
            return self
        if isinstance(other, self.__class__):
            if _lib.Add(self._buffer, self._contents, other._contents):
                raise ArithmeticError, "Patterns are not disjoint"
            result = self._wrapbuffer()
#            if self.isPeriodic and self._pst == other._pst:
#                try: result.setPeriod(self.period)
#                except ValueError: pass
            if self.isPeriodic and other.isPeriodic:
                p, q = self.period, other.period
                try: result.setPeriod(p * q / gcd(p, q))
                except ValueError: pass
            return result
        return NotImplemented

    def __radd__(self, other):
        return self.__add__(other)
    
    def __rmul__(self, other):
        if isinstance(other, Xform):
            return self.xformed(other)
        return NotImplemented
    
    def __div__(self, other):
        if isinstance(other, Xform):
            return self.__rmul__(~other)
        if isinstance(other, self.__class__):
            if self.homed == other.homed:
                return other.homer / self.homer
            elif (self.isPeriodic and other.isPeriodic and
                  self.period == other.period):
                p = self.homed
                for xf in X.ORIENTATIONS:
                    q = xf * other
                    for t in range(self.period):
                        if p == q[t].homed:
                            return (self / q[t]) * xf * X(t)
            raise ValueError, "Pattern division is too hard"
        return NotImplemented
    
#    def __eq__(self, other):
#        if isinstance(other, self.__class__):
#            return _lib.Compare(self._contents, other._contents) == 0
#        return False
#    
#    def __ne__(self, other):
#        return not self.__eq__(other)
    
    def __cmp__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        result = cmp(self.pop, other.pop)
        if result != 0: return result
        return (result if result != 0 else
                cmp(self._rawData, other._rawData))

    def _boxes(self):
        if self._header.flags & _BOUNDINGBOX_MASK == 0:
            _lib.BoundingBox(self._contents)
        return self._header.boxes

    def _jogged(self):
        _lib.Jog(self._buffer, self._contents)
        return self._wrapbuffer()

    def divine(self, n):
        if self.isPeriodic :
            return n % self.period == 0
        for i in self._factors(n):
            try:
                self.setPeriod(i, verify=True)
                return True
            except ValueError:
                pass
        return False

    def translated(self, x, y):
        if x == y == 0: return self
        _lib.Translate(self._buffer, self._contents,
                       c_int(x), c_int(y))
        return self._wrapbuffer()
    
    def xformed(self, xform):
        if xform.isReoriented:
            t = xform.t
            xform /= X(t)
            cells = list(self[t].cells)
            for i in range(0, len(cells), 2):
                cells[i:i+2] = xform(cells[i:i+2])
            result = self.__class__(cells)
            if self.isPeriodic:
                result.setPeriod(self.period)
            return result
        else:
            return self[xform.t].translated(*xform.xy)
        
    def smudged(self, n=1):
        _lib.Smudge(self._buffer, self._contents, c_int(n))
        return self._wrapbuffer()

    def smear(self, n=1):
        _lib.Gen(self._buffer, self._contents, self._smearBuffer,
                 c_int(n), None, None)
        return (self._wrapbuffer(), self._wrapsmear())

    def monitor(self, n, smear=False):
        sb = self._smearBuffer if smear else None
        rc = c_int(0)
        n = _lib.Gen(self._buffer, self._contents, sb, c_int(n), None,
                     byref(rc))
        if smear:
            return (self._wrapbuffer(), n, rc.value, self._wrapsmear())
        else:
            return (self._wrapbuffer(), n, rc.value)
        
    def setPeriod(self, n, x=None, y=None, verify=True):
        header = self._header
        if header.flags & _PERIODIC_MASK:
            if n != self.period or (x != None and y != None and
                                    (x, y) != self.stride):
                raise ValueError, "Attempt to change period/stride"
            return
        if n <= 0:
            raise ValueError, "Period must be positive"
        p = self[n] if verify else None
        if x == None or y == None: 
            if verify:
                try: x, y = (p / self).xy
                except ValueError:
                    raise ValueError, "Period verification failed"
            else:
                x = y = 0
        else:
            if verify and p != self.translated(x, y):
                raise ValueError, "Period/stride verification failed"
        header.period, header.xstride, header.ystride = n, x, y
        header.flags |= _PERIODIC_MASK
        
    def census(self, n):
        result = []
        p = self
        while p:
            q = self.__class__(*p.firstcell)
            while True:
                r = q
                q = q.smudged(n) & p
                if q == r: break
            result += [q]
            p -= q
        return result
    
    def blocksNWSalvo(self, salvo):
        delay = 2 * (self.dbox[1] + self.dbox[3] - salvo.dbox[1]) + 8
        safeSalvo = salvo[-delay]
        t = 2 * (safeSalvo.dbox[1] + safeSalvo.dbox[3] -
                 self.dbox[1]) + 4
        return (self | safeSalvo)[t] != self[t] | safeSalvo[t]
    
P = Pattern

P.E0 = P.EMPTY   = E0 = EMPTY   = P()
P.B3 = P.BLINKER = B3 = BLINKER = P("3o!")
P.B4 = P.BLOCK   = B4 = BLOCK   = P("2o$2o!")
P.T4 = P.TUB     = T4 = TUB     = P("bo$obo$bo!")
P.B5 = P.BOAT    = B5 = BOAT    = P("2o$obo$bo!")
P.B6 = P.BEEHIVE = B6 = BEEHIVE = P("b2o$o2bo$b2o!")

P.G  = P.GLIDER  = G  = GLIDER  = P("3o$o$bo!")

P.PATTERNS = ["EMPTY", "BLINKER", "BLOCK", "TUB", "BOAT", "BEEHIVE",
              "GLIDER"]

for p in [E0, B4, T4, B5, B6]: p.setPeriod(1)
for p in [B3]: p.setPeriod(2)
for p in [G]: p.setPeriod(4)

def GX(x, t): return G[t].translated(x, 0)
def GX2(lane, d): return G[2 * lane - d].translated(lane, 0)

def StampCollection(c):
    maxw = maxh = 0
    for p in c:
        w, h = p.box[2:]
        maxw, maxh = max(maxw, w), max(maxh, h)
    for i in range(len(c)): p |= c[i] 
    w, h = p.box[2:]
    result = E0
    x = 0
    for p in c:
        result = result + X(x, 0) * p
        x += maxw + 6
    return result

__all__ = ["Pattern", "P", "Xform", "X",
           "GX", "GX2",
           "XI", "XR", "XC", "XA", "XH", "XV", "XD", "XS",
           "EMPTY", "BLINKER", "BLOCK", "TUB", "BOAT", "BEEHIVE",
           "GLIDER",
           "E0", "B3", "B4", "T4", "B5", "B6", "G"]

if __name__ == "__main__":
    print BLINKER.canonicalPhase.period
    print BLINKER[1].canonicalPhase.period
    print BLINKER[2].canonicalPhase.period
    
#    p = P(0, 0)
#    print p.__class__()
#    print p._contents[:]
#    print p._header.size
#    print p.pop
#    print p.__hash__()
#    print p._jogged()
#    print p._jogged().__hash__()
#    
#    q = p.smudged()
#    print q
#    print q.cells
#    print q.firstcell
#    
#    print G.cells
#    print (X(1, 1) * G[4]).cells
#    print G._header.flags
#    print G.box
#    print G._header.flags
#    print repr(X(10, 10) * G.smear(4)[1])
#    print G.monitor(100)
#
#    print repr(G)
#    print repr(X(1, 1) * G)
#    print repr(X(1, 1) * G / X(1, 1))
#
#    print G.period
#    print G.stride
#    print G.box
#    print G[4].box
#    print G[4] / G
#    
#    print XI
#    print X(3)
#    print XI.isReoriented
#    print XR.isReoriented
#    
#    p = XA * G[1]
#    print G, p
#    print p / G
#    print p / G * G
